import { BatchResult } from 'src/core/response/domain/ok-response.interface';
import { BaseManager } from '../base.manager';
import { HttpStatus, NotFoundException } from '@nestjs/common';
import { OPERATION, STATUS } from 'src/core/strings/constants/base.constants';
import { ValidateRelationHelper } from 'src/core/helpers/validation/validate-relation.helper';
import { RecordLog } from 'src/modules/configuration/log/domain/entities/log.event';
import * as _ from 'lodash';

export abstract class BaseBatchUpdateStatusManager<Entity> extends BaseManager {
  protected dataIds: string[];
  protected relations: string[] = [];
  protected result: BatchResult;
  protected dataStatus: STATUS;
  protected oldData: Entity;
  abstract get entityTarget(): any;

  setData(ids: string[], status: STATUS): void {
    this.dataIds = ids;
    this.dataStatus = status;
  }

  validateProcess(): Promise<void> {
    return;
  }

  prepareData(): Promise<void> {
    return;
  }

  async process(): Promise<void> {
    let totalFailed = 0;
    let totalSuccess = 0;
    const messages = [];

    for (const id of this.dataIds) {
      try {
        const entity = await this.dataService.getOneByOptions({
          where: {
            id: id,
          },
          relations: this.relations,
        });

        if (!entity) {
          throw new NotFoundException({
            statusCode: HttpStatus.NOT_FOUND,
            message: `Failed! Entity with id ${id} not found`,
            error: 'Entity Not Found',
          });
        }
        this.oldData = _.cloneDeep(entity);
        await this.validateData(entity);

        Object.assign(entity, {
          status: this.dataStatus,
          editor_id: this.user.id,
          editor_name: this.user.name,
          updated_at: new Date().getTime(),
        });
        await new ValidateRelationHelper(
          id,
          this.dataService,
          this.validateRelations,
          this.tableName,
        ).execute();

        const result = await this.dataService.update(
          this.queryRunner,
          this.entityTarget,
          { id: id },
          entity,
        );

        this.publishEvents(this.oldData, result);

        totalSuccess = totalSuccess + 1;
      } catch (error) {
        totalFailed = totalFailed + 1;
        messages.push(error.response?.message ?? error.message);
      }
    }

    this.result = {
      total_items: this.dataIds.length,
      total_failed: totalFailed,
      total_success: totalSuccess,
      messages: messages,
    };
  }

  abstract validateData(data: Entity): Promise<void>;
  abstract getResult(): BatchResult;

  async publishEvents(dataOld, dataNew) {
    this.eventBus.publish(
      new RecordLog({
        id: dataNew['id'],
        old: dataOld,
        data: dataNew,
        user: this.user,
        description: `${this.user.name} update batch data ${this.tableName}`,
        module: this.tableName,
        op: OPERATION.UPDATE,
      }),
    );

    if (!this.eventTopics.length) return;
    for (const topic of this.eventTopics) {
      let data;
      if (topic.relations?.length) {
        data = await this.dataService.getOneByOptions({
          where: {
            id: dataNew.id,
          },
          relations: topic.relations,
        });
      }

      this.eventBus.publishAll([
        new topic.topic({
          id: data?.['id'] ?? dataNew?.['id'],
          old: dataOld,
          data: data ?? dataNew,
          user: this.user,
          description: '',
          module: this.tableName,
          op: OPERATION.UPDATE,
        }),
      ]);
    }
  }
}
